"""Datasets built from music21 corpus."""
from pathlib import Path
from typing import Union

from music21 import corpus

from ..inputs import read
from ..music import Music
from .base import Dataset, DatasetInfo, MusicDataset

_NAME = "Music21 Corpus Dataset"
_DESCRIPTION = """Dataset automatically created from a music21 corpus."""
_HOMEPAGE = "https://web.mit.edu/music21/doc/about/referenceCorpus.html"
_CITATION = """\
@inproceedings{cuthbert2010music21,
  author={Michael Scott Cuthbert and Christopher Ariza},
  title={Music21: A Toolkit for Computer-Aided Musicology and Symbolic Music Data},
  booktitle={Proceedings of the 11th International Society for Music Information Retrieval Conference (ISMIR)},
  year=2010
}"""


class Music21Dataset(Dataset):
    """A class of datasets containing files in music21 corpus.

    Parameters
    ----------
    composer : str
        Name of a composer or a collection. Please refer to the music21
        corpus reference page for a full list [1].
    extensions : list of str
        File extensions of desired files.

    References
    ----------
    [1] https://web.mit.edu/music21/doc/about/referenceCorpus.html

    """

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _citation = _CITATION
    _extensions = (
        ".mid",
        ".midi",
        ".mxl",
        ".xml",
        ".mxml",
        ".musicxml",
        # ".abc",
    )

    def __init__(self, composer: str = None):
        if composer is None:
            self.composer = "ALL"
            self.filenames = [
                path
                for path in corpus.corpora.CoreCorpus().getPaths()
                if str(path).endswith(self._extensions)
            ]
        else:
            self.composer = composer
            self.filenames = corpus.getComposer(composer, self._extensions)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(composer={self.composer})"

    def __getitem__(self, index) -> Music:
        if str(self.filenames[index]).lower().endswith(".abc"):
            return read(self.filenames[index], number=0)  # type: ignore
        return read(self.filenames[index])  # type: ignore

    def __len__(self) -> int:
        return len(self.filenames)

    def convert(
        self,
        root: Union[str, Path],
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
    ) -> "MusicDataset":
        """Convert and save the Music objects.

        Parameters
        ----------
        root : str or Path
            Root directory to save the data.
        kind : {'json', 'yaml'}, default: 'json'
            File format to save the data.
        n_jobs : int, default: 1
            Maximum number of concurrently running jobs. If equal to 1,
            disable multiprocessing.
        ignore_exceptions : bool, default: True
            Whether to ignore errors and skip failed conversions. This
            can be helpful if some source files are known to be
            corrupted.

        """
        self.save(root, kind, n_jobs, ignore_exceptions)
        return MusicDataset(root, kind)



"""YAML output interface."""
import gzip
from pathlib import Path
from typing import TYPE_CHECKING, TextIO, Union

from ..utils import yaml_dump

if TYPE_CHECKING:
    from ..music import Music


def save_yaml(
    path: Union[str, Path, TextIO],
    music: "Music",
    skip_missing: bool = True,
    allow_unicode: bool = True,
    compressed: bool = None,
    **kwargs
):
    """Save a Music object to a YAML file.

    Parameters
    ----------
    path : str, Path or TextIO
        Path or file to save the YAML data.
    music : :class:`muspy.Music`
        Music object to save.
    skip_missing : bool, default: True
        Whether to skip attributes with value None or those that are
        empty lists.
    allow_unicode : bool, default: False
        Whether to escape non-ASCII characters. Will be passed to
        :py:func:`json.dumps`.
    compressed : bool, optional
        Whether to save as a compressed YAML file (`.yaml.gz`). Has no
        effect when `path` is a file object. Defaults to infer from the
        extension (`.gz`).
    **kwargs
        Keyword arguments to pass to `yaml.dump`.

    Notes
    -----
    When a path is given, use UTF-8 encoding and gzip compression if
    `compressed=True`.

    """
    data = yaml_dump(
        music.to_ordered_dict(skip_missing=skip_missing, deepcopy=False),
        allow_unicode=allow_unicode,
        **kwargs
    )

    if isinstance(path, (str, Path)):
        if compressed is None:
            if str(path).lower().endswith(".gz"):
                compressed = True
            else:
                compressed = False
        if compressed:
            with gzip.open(path, "wt", encoding="utf-8") as f:
                f.write(data)
        else:
            with open(path, "w", encoding="utf-8") as f:
                f.write(data)
        return

    path.write(data)



"""MIDI output interface."""
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Tuple, Union

import numpy as np
from mido import Message, MetaMessage, MidiFile, MidiTrack, bpm2tempo
from pretty_midi import Instrument
from pretty_midi import KeySignature as PmKeySignature
from pretty_midi import Lyric as PmLyric
from pretty_midi import Note as PmNote
from pretty_midi import PrettyMIDI
from pretty_midi import TimeSignature as PmTimeSignature
from pretty_midi import key_name_to_key_number

from ..classes import (
    DEFAULT_VELOCITY,
    KeySignature,
    Lyric,
    Note,
    Tempo,
    TimeSignature,
    Track,
)

if TYPE_CHECKING:
    from ..music import Music

PITCH_NAMES = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
DEFAULT_TEMPO = 120


def to_delta_time(midi_track: MidiTrack):
    """Convert a mido MidiTrack object from absolute time to delta time.

    Parameters
    ----------
    midi_track : :class:`mido.MidiTrack` object
        mido MidiTrack object to convert.

    """
    # Sort messages by absolute time
    midi_track.sort(key=lambda x: x.time)

    # Convert to delta time
    time = 0
    for msg in midi_track:
        time_ = msg.time
        msg.time -= time
        time = time_


def to_mido_tempo(tempo: Tempo) -> MetaMessage:
    """Return a Tempo object as a mido MetaMessage object.

    Timing is in absolute time, NOT in delta time.

    """
    return MetaMessage(
        "set_tempo", time=tempo.time, tempo=bpm2tempo(tempo.qpm),
    )


def to_mido_key_signature(
    key_signature: KeySignature,
) -> Optional[MetaMessage]:
    """Return a KeySignature object as a mido MetaMessage object.

    Timing is in absolute time, NOT in delta time.

    """
    # TODO: `key_signature.root_str` might be given
    if key_signature.root is None:
        return None
    if key_signature.mode not in ("major", "minor"):
        return None
    suffix = "m" if key_signature.mode == "minor" else ""
    return MetaMessage(
        "key_signature",
        time=key_signature.time,
        key=PITCH_NAMES[key_signature.root] + suffix,
    )


def to_mido_time_signature(time_signature: TimeSignature) -> MetaMessage:
    """Return a TimeSignature object as a mido MetaMessage object.

    Timing is in absolute time, NOT in delta time.

    """
    return MetaMessage(
        "time_signature",
        time=time_signature.time,
        numerator=time_signature.numerator,
        denominator=time_signature.denominator,
    )


def to_mido_meta_track(music: "Music") -> MidiTrack:
    """Return a mido MidiTrack containing metadata of a Music object.

    Parameters
    ----------
    music : :class:`muspy.Music` object
        Music object to convert.

    Returns
    -------
    :class:`mido.MidiTrack` object
        Converted mido MidiTrack object.

    """
    # Create a track to store the metadata
    meta_track = MidiTrack()

    # Song title
    if music.metadata.title is not None:
        meta_track.append(MetaMessage("track_name", name=music.metadata.title))

    # Tempos
    for tempo in music.tempos:
        meta_track.append(to_mido_tempo(tempo))

    # Key signatures
    for key_signature in music.key_signatures:
        mido_key_signature = to_mido_key_signature(key_signature)
        if mido_key_signature is not None:
            meta_track.append(mido_key_signature)

    # Time signatures
    for time_signature in music.time_signatures:
        meta_track.append(to_mido_time_signature(time_signature))

    # Lyrics
    for lyric in music.lyrics:
        meta_track.append(to_mido_lyric(lyric))

    # Annotations
    for annotation in music.annotations:
        # Marker messages
        if annotation.group == "marker":
            meta_track.append(
                MetaMessage("marker", text=annotation.annotation)
            )
        # Text messages
        elif isinstance(annotation.annotation, str):
            meta_track.append(
                MetaMessage(
                    "text", time=annotation.time, text=annotation.annotation
                )
            )

    # End of track message
    meta_track.append(MetaMessage("end_of_track"))

    # Convert to delta time
    to_delta_time(meta_track)

    return meta_track


def to_mido_lyric(lyric: Lyric) -> MetaMessage:
    """Return a Lyric object as a mido MetaMessage object.

    Timing is in absolute time, NOT in delta time.

    """
    return MetaMessage("lyrics", time=lyric.time, text=lyric.lyric)


def to_mido_note_on_note_off(
    note: Note, channel: int, use_note_off_message: bool = False
) -> Tuple[Message, Message]:
    """Return a Note object as mido Message objects.

    Timing is in absolute time, NOT in delta time.

    Parameters
    ----------
    note : :class:`muspy.Note` object
        Note object to convert.
    channel : int
        Channel of the MIDI message.
    use_note_off_message : bool, default: False
        Whether to use note-off messages. If False, note-on messages
        with zero velocity are used instead. The advantage to using
        note-on messages at zero velocity is that it can avoid sending
        additional status bytes when Running Status is employed.

    Returns
    -------
    :class:`mido.Message` object
        Converted mido Message object for note on.
    :class:`mido.Message` object
        Converted mido Message object for note off.

    """
    velocity = note.velocity if note.velocity is not None else DEFAULT_VELOCITY
    note_on_msg = Message(
        "note_on",
        time=note.time,
        note=note.pitch,
        velocity=velocity,
        channel=channel,
    )
    if use_note_off_message:
        note_off_msg = Message(
            "note_off",
            time=note.end,
            note=note.pitch,
            velocity=64,
            channel=channel,
        )
    else:
        note_off_msg = Message(
            "note_on",
            time=note.end,
            note=note.pitch,
            velocity=0,
            channel=channel,
        )

    return note_on_msg, note_off_msg


def to_mido_track(
    track: Track, channel: int = None, use_note_off_message: bool = False,
) -> MidiTrack:
    """Return a Track object as a mido MidiTrack object.

    Parameters
    ----------
    track : :class:`muspy.Track` object
        Track object to convert.
    channel : int, optional
        Channel number. Defaults to 10 for drums and 0 for other
        instruments.
    use_note_off_message : bool, default: False
        Whether to use note-off messages. If False, note-on messages
        with zero velocity are used instead. The advantage to using
        note-on messages at zero velocity is that it can avoid sending
        additional status bytes when Running Status is employed.

    Returns
    -------
    :class:`mido.MidiTrack` object
        Converted mido MidiTrack object.

    """
    if channel is None:
        channel = 9 if track.is_drum else 0

    # Create a new MIDI track
    midi_track = MidiTrack()

    # Track name messages
    if track.name is not None:
        midi_track.append(MetaMessage("track_name", name=track.name))

    # Program change messages
    midi_track.append(
        Message("program_change", program=track.program, channel=channel)
    )

    # Note on and note off messages
    for note in track.notes:
        midi_track.extend(
            to_mido_note_on_note_off(
                note,
                channel=channel,
                use_note_off_message=use_note_off_message,
            )
        )

    # End of track message
    midi_track.append(MetaMessage("end_of_track"))

    # Convert to delta time
    to_delta_time(midi_track)

    return midi_track


def to_mido(music: "Music", use_note_off_message: bool = False):
    """Return a Music object as a MidiFile object.

    Parameters
    ----------
    music : :class:`muspy.Music` object
        Music object to convert.
    use_note_off_message : bool, default: False
        Whether to use note-off messages. If False, note-on messages
        with zero velocity are used instead. The advantage to using
        note-on messages at zero velocity is that it can avoid sending
        additional status bytes when Running Status is employed.

    Returns
    -------
    :class:`mido.MidiFile`
        Converted MidiFile object.

    """
    # Create a MIDI file object
    midi = MidiFile(type=1, ticks_per_beat=music.resolution)

    # Append meta track
    midi.tracks.append(to_mido_meta_track(music))

    # Iterate over music tracks
    for i, track in enumerate(music.tracks):
        # NOTE: Many softwares use the same instrument for messages of
        # the same channel in different tracks. Thus, we want to assign
        # a unique channel number for each track. MIDI has 15 channels
        # for instruments other than drums, so we increment the channel
        # number for each track (skipping the drum channel) and go back
        # to 0 once we run out of channels.

        # Assign channel number
        if track.is_drum:
            # Mido numbers channels 0 to 15 instead of 1 to 16
            channel = 9
        else:
            # MIDI has 15 channels for instruments other than drums
            channel = i % 15
            # Avoid drum channel
            if channel > 8:
                channel += 1

        midi.tracks.append(
            to_mido_track(
                track,
                channel=channel,
                use_note_off_message=use_note_off_message,
            )
        )

    return midi


def write_midi_mido(
    path: Union[str, Path], music: "Music", use_note_off_message: bool = False,
):
    """Write a Music object to a MIDI file using mido as backend.

    Parameters
    ----------
    path : str or Path
        Path to write the MIDI file.
    music : :class:`muspy.Music` object
        Music object to write.
    use_note_off_message : bool, default: False
        Whether to use note-off messages. If False, note-on messages
        with zero velocity are used instead. The advantage to using
        note-on messages at zero velocity is that it can avoid sending
        additional status bytes when Running Status is employed.

    """
    midi = to_mido(music, use_note_off_message=use_note_off_message)
    midi.save(str(path))


def to_pretty_midi_key_signature(
    key_signature: KeySignature,
) -> Optional[PmKeySignature]:
    """Return a KeySignature object as a pretty_midi KeySignature."""
    # TODO: `key_signature.root_str` might be given
    if key_signature.root is None:
        return None
    if key_signature.mode not in ("major", "minor"):
        return None
    key_name = PITCH_NAMES[key_signature.root] + " " + key_signature.mode
    return PmKeySignature(
        key_number=key_name_to_key_number(key_name), time=key_signature.time,
    )


def to_pretty_midi_time_signature(
    time_signature: TimeSignature,
) -> PmTimeSignature:
    """Return a TimeSignature object as a pretty_midi TimeSignature."""
    return PmTimeSignature(
        numerator=time_signature.numerator,
        denominator=time_signature.denominator,
        time=time_signature.time,
    )


def to_pretty_midi_lyric(lyric: Lyric) -> PmLyric:
    """Return a Lyric object as a pretty_midi Lyric object."""
    return PmLyric(lyric.lyric, lyric.time)


def to_pretty_midi_note(note: Note) -> PmNote:
    """Return a Note object as a pretty_midi Note object."""
    velocity = note.velocity if note.velocity is not None else DEFAULT_VELOCITY
    return PmNote(
        velocity=velocity, pitch=note.pitch, start=note.time, end=note.end
    )


def to_pretty_midi_instrument(track: Track) -> Instrument:
    """Return a Track object as a pretty_midi Instrument object."""
    instrument = Instrument(
        program=track.program, is_drum=track.is_drum, name=track.name
    )
    for note in track.notes:
        instrument.notes.append(to_pretty_midi_note(note))
    return instrument


def to_pretty_midi(music: "Music") -> PrettyMIDI:
    """Return a Music object as a PrettyMIDI object.

    Tempo changes are not supported yet.

    Parameters
    ----------
    music : :class:`muspy.Music` object
        Music object to convert.

    Returns
    -------
    :class:`pretty_midi.PrettyMIDI`
        Converted PrettyMIDI object.

    Notes
    -----
    Tempo information will not be included in the output.

    """
    # Create an PrettyMIDI instance
    midi = PrettyMIDI()

    # Compute tempos
    tempo_times, tempi = [0], [float(DEFAULT_TEMPO)]
    for tempo in music.tempos:
        tempo_times.append(tempo.time)
        tempi.append(tempo.qpm)

    # Remove unnecessary tempo changes to speed up the search
    if len(tempi) > 1:
        last_tempo = tempi[0]
        last_time = tempo_times[0]
        i = 1
        while i < len(tempo_times):
            if tempi[i] == last_tempo:
                del tempo_times[i]
                del tempi[i]
            elif tempo_times[i] == last_time:
                del tempo_times[i - 1]
                del tempi[i - 1]
            else:
                last_tempo = tempi[i]
                i += 1
        tempo_times = np.array(tempo_times)
        tempi = np.array(tempi)

    if len(tempi) == 1:

        def map_time(time):
            return time * 60.0 / (music.resolution * tempi[0])

    else:
        # Compute the tempo time in absolute timing of each tempo change
        tempo_realtimes = np.cumsum(
            np.diff(tempo_times) * 60.0 / (music.resolution * tempi[:-1])
        ).tolist()
        tempo_realtimes.insert(0, 0.0)

        def map_time(time):
            idx = np.searchsorted(tempo_times, time, side="right") - 1
            residual = time - tempo_times[idx]
            factor = 60.0 / (music.resolution * tempi[idx])
            return tempo_realtimes[idx] + residual * factor

    # Key signatures
    for key_signature in music.key_signatures:
        pm_key_signature = to_pretty_midi_key_signature(key_signature)
        if pm_key_signature is not None:
            pm_key_signature.time = map_time(pm_key_signature.time)
            midi.key_signature_changes.append(pm_key_signature)

    # Time signatures
    for time_signature in music.time_signatures:
        midi.time_signature_changes.append(
            to_pretty_midi_time_signature(time_signature)
        )

    # Lyrics
    for lyric in music.lyrics:
        midi.lyrics.append(to_pretty_midi_lyric(lyric))

    # Tracks
    for track in music.tracks:
        midi.instruments.append(to_pretty_midi_instrument(track))

    return midi


def write_midi_pretty_midi(path: Union[str, Path], music: "Music"):
    """Write a Music object to a MIDI file using pretty_midi as backend.

    Tempo changes are not supported yet.

    Parameters
    ----------
    path : str or Path
        Path to write the MIDI file.
    music : :class:`muspy.Music` object
        Music object to convert.

    Notes
    -----
    Tempo information will not be included in the output.

    """
    midi = to_pretty_midi(music)
    midi.write(str(path))


def write_midi(
    path: Union[str, Path],
    music: "Music",
    backend: str = "mido",
    **kwargs: Any
):
    """Write a Music object to a MIDI file.

    Parameters
    ----------
    path : str or Path
        Path to write the MIDI file.
    music : :class:`muspy.Music`
        Music object to write.
    backend: {'mido', 'pretty_midi'}, default: 'mido'
        Backend to use.

    See Also
    --------
    write_midi_mido :
        Write a Music object to a MIDI file using mido as backend.
    write_midi_pretty_midi :
        Write a Music object to a MIDI file using pretty_midi as
        backend.

    """
    if backend == "mido":
        return write_midi_mido(path, music, **kwargs)
    if backend == "pretty_midi":
        return write_midi_pretty_midi(path, music)
    raise ValueError("`backend` must by one of 'mido' and 'pretty_midi'.")



"""Base Dataset classes."""
import json
import warnings
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
from joblib import Parallel, delayed
from numpy.random import RandomState, permutation
from tqdm import tqdm

from ..inputs import load, read_abc_string
from ..music import Music
from ..outputs import save
from .utils import (
    check_md5,
    check_sha256,
    check_size,
    download_url,
    extract_archive,
)

if TYPE_CHECKING:
    from tensorflow.data import Dataset as TFDataset
    from torch.utils.data import Dataset as TorchDataset


RemoteDatasetType = TypeVar("RemoteDatasetType", bound="RemoteDataset")
FolderDatasetType = TypeVar("FolderDatasetType", bound="FolderDataset")


class DatasetInfo:
    """A container for dataset information."""

    def __init__(
        self,
        name: str = None,
        description: str = None,
        homepage: str = None,
        license: str = None,
    ):
        # pylint: disable=redefined-builtin
        self.name = name
        self.description = description
        self.homepage = homepage
        self.license = license

    def __repr__(self) -> str:
        to_join = []
        for attr in ("name", "description", "homepage", "license"):
            if getattr(self, attr) is not None:
                to_join.append(attr + "=" + repr(getattr(self, attr)))
        return "DatasetInfo(" + ", ".join(to_join) + ")"


class Dataset:
    """Base class for MusPy datasets.

    To build a custom dataset, it should inherit this class and overide
    the methods ``__getitem__`` and ``__len__`` as well as the class
    attribute ``_info``. ``__getitem__`` should return the ``i``-th data
    sample as a :class:`muspy.Music` object. ``__len__`` should return
    the size of the dataset. ``_info`` should be a
    :class:`muspy.DatasetInfo` instance storing the dataset information.

    """

    _info: DatasetInfo = DatasetInfo()
    _citation: str = ""

    def __getitem__(self, index) -> Music:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    @classmethod
    def info(cls):
        """Return the dataset infomation."""
        return cls._info

    @classmethod
    def citation(cls):
        """Print the citation infomation."""
        return cls._citation

    def save(
        self,
        root: Union[str, Path],
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        verbose: bool = True,
        **kwargs,
    ):
        """Save all the music objects to a directory.

        Parameters
        ----------
        root : str or Path
            Root directory to save the data.
        kind : {'json', 'yaml'}, default: 'json'
            File format to save the data.
        n_jobs : int, default: 1
            Maximum number of concurrently running jobs. If equal to 1,
            disable multiprocessing.
        ignore_exceptions : bool, default: True
            Whether to ignore errors and skip failed conversions. This
            can be helpful if some source files are known to be
            corrupted.
        verbose : bool, default: True
            Whether to be verbose.
        **kwargs
            Keyword arguments to pass to :func:`muspy.save`.

        """
        if kind not in ("json", "yaml"):
            raise TypeError("`kind` must be either 'json' or 'yaml'.")

        root = Path(root).expanduser().resolve()
        root.mkdir(exist_ok=True)

        def _saver(idx):
            prefix = "0" * (n_digits - len(str(idx)))
            filename = root / (prefix + str(idx) + "." + kind)
            if ignore_exceptions:
                try:
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        save(filename, self[idx], kind, **kwargs)
                except Exception:  # pylint: disable=broad-except
                    return False
                return True
            save(filename, self[idx], kind, **kwargs)
            return True

        n_digits = len(str(len(self)))

        if verbose:
            print("Converting and saving the dataset...")
        if n_jobs == 1:
            count = 0
            for idx in tqdm(range(len(self))):  # type: ignore
                if _saver(idx):
                    count += 1
        else:
            # TODO: This is slow as `self` is passed between workers.
            results = Parallel(n_jobs=n_jobs, backend="threading", verbose=5)(
                delayed(_saver)(idx) for idx in range(len(self))
            )
            count = results.count(True)
        if verbose:
            print(f"Successfully saved {count} out of {len(self)} files.")

    def split(
        self,
        filename: Union[str, Path] = None,
        splits: Sequence[float] = None,
        random_state: Any = None,
    ) -> Dict[str, List[int]]:
        """Return the dataset as a PyTorch dataset.

        Parameters
        ----------
        filename : str or Path, optional
            If given and exists, path to the file to read the split
            from. If None or not exists, path to save the split.
        splits : float or list of float, optional
            Ratios for train-test-validation splits. If None, return the
            full dataset as a whole. If float, return train and test
            splits. If list of two floats, return train and test splits.
            If list of three floats, return train, test and validation
            splits.
        random_state : int, array_like or RandomState, optional
            Random state used to create the splits. If int or
            array_like, the value is passed to
            :class:`numpy.random.RandomState`, and the created
            RandomState object is used to create the splits. If
            RandomState, it will be used to create the splits.

        """
        if filename is not None and Path(filename).is_file():
            with open(str(filename)) as f:
                return json.load(f)

        if not isinstance(splits, (float, list, tuple)):
            raise TypeError("`splits` must be of type float, list or tuple.")

        if isinstance(splits, float):
            if splits <= 0:
                raise ValueError("`splits` must be positive.")
            if splits >= 1:
                raise ValueError("`splits` must be less than 1.")
            splits = [splits, 1 - splits]

        if isinstance(splits, (list, tuple)):
            if sum(splits) != 1:
                raise ValueError("`splits` must sum to 1.")
            if len(splits) < 2 or len(splits) > 3:
                raise ValueError("`splits` must have length 2 or 3.")

        if random_state is None:
            rand_indices = permutation(len(self))
        else:
            if not isinstance(random_state, RandomState):
                random_state = RandomState(random_state)
            rand_indices = random_state.permutation(len(self))

        boundaries = np.cumsum([0.0] + list(splits))
        names = ("train", "test", "validation")
        indices = {}
        for idx, (start, end) in enumerate(
            zip(boundaries[:-1], boundaries[1:])
        ):
            start_idx = int(start * len(self))
            end_idx = int(end * len(self))
            indices[names[idx]] = rand_indices[start_idx:end_idx]

        if filename is not None:
            indices_ = {key: value.tolist() for key, value in indices.items()}
            with open(str(filename), "w") as f:
                f.write(json.dumps(indices_))

        return indices

    def to_pytorch_dataset(
        self,
        factory: Callable = None,
        representation: str = None,
        split_filename: Union[str, Path] = None,
        splits: Sequence[float] = None,
        random_state: Any = None,
        **kwargs: Any,
    ) -> Union["TorchDataset", Dict[str, "TorchDataset"]]:
        """Return the dataset as a PyTorch dataset.

        Parameters
        ----------
        factory : Callable, optional
            Function to be applied to the Music objects. The input is a
            Music object, and the output is an array or a tensor.
        representation : str, optional
            Target representation. See :func:`muspy.to_representation()`
            for available representation.
        split_filename : str or Path, optional
            If given and exists, path to the file to read the split
            from. If None or not exists, path to save the split.
        splits : float or list of float, optional
            Ratios for train-test-validation splits. If None, return the
            full dataset as a whole. If float, return train and test
            splits. If list of two floats, return train and test splits.
            If list of three floats, return train, test and validation
            splits.
        random_state : int, array_like or RandomState, optional
            Random state used to create the splits. If int or
            array_like, the value is passed to
            :class:`numpy.random.RandomState`, and the created
            RandomState object is used to create the splits. If
            RandomState, it will be used to create the splits.

        Returns
        -------
        :class:torch.utils.data.Dataset` or Dict of \
                :class:torch.utils.data.Dataset`
            Converted PyTorch dataset(s).

        """
        if representation is None and factory is None:
            raise TypeError(
                "One of `representation` and `factory` must be given."
            )
        if representation is not None and factory is not None:
            raise TypeError(
                "Only one of `representation` and `factory` can be given."
            )

        try:
            from torch.utils.data import Dataset as TorchDataset
        except ImportError:
            raise ImportError("Optional package pytorch is required.")

        class TorchMusicFactoryDataset(TorchDataset):
            """A PyTorch dataset built from a Music dataset.

            Parameters
            ----------
            dataset : :class:`muspy.Dataset`
                Dataset object to base on.
            factory : Callable
                Function to be applied to the Music objects. The input is a
                Music object, and the output is an array or a tensor.

            """

            def __init__(
                self,
                dataset: Dataset,
                factory: Callable,
                subset: str = "Full",
                indices: Sequence[int] = None,
            ):
                self.dataset = dataset
                self.factory = factory
                self.subset = subset
                self.indices = indices
                if self.indices is not None:
                    self.indices = sorted(
                        idx for idx in self.indices if idx < len(self.dataset)
                    )

            def __repr__(self) -> str:
                return (
                    f"TorchMusicFactoryDataset(dataset={self.dataset}, "
                    f"factory={self.subset}, subset={self.factory})"
                )

            def __getitem__(self, index):
                if self.indices is None:
                    return self.factory(self.dataset[index])
                return self.factory(self.dataset[self.indices[index]])

            def __len__(self) -> int:
                if self.indices is None:
                    return len(self.dataset)
                return len(self.indices)

        class TorchRepresentationDataset(TorchMusicFactoryDataset):
            """A PyTorch music dataset.

            Parameters
            ----------
            dataset : :class:`muspy.Dataset`
                Dataset object to base on.
            representation : str
                Target representation. See
                :func:`muspy.to_representation()` for available
                representation.

            """

            def __init__(
                self,
                dataset: Dataset,
                representation: str,
                subset: str = "Full",
                indices: Sequence[int] = None,
                **kwargs: Any,
            ):
                self.representation = representation

                def factory(music):
                    return music.to_representation(representation, **kwargs)

                super().__init__(
                    dataset, factory=factory, subset=subset, indices=indices
                )

            def __repr__(self) -> str:
                return (
                    f"TorchRepresentationDataset(dataset={self.dataset}, "
                    f"representation={self.representation}, "
                    f"subset={self.subset})"
                )

        # No split
        if splits is None:
            if representation is not None:
                return TorchRepresentationDataset(
                    self, representation, **kwargs
                )
            return TorchMusicFactoryDataset(self, factory)  # type: ignore

        datasets: Dict[str, "TorchDataset"] = {}
        indices_list = self.split(split_filename, splits, random_state)
        for key, value in indices_list.items():
            if representation is not None:
                datasets[key] = TorchRepresentationDataset(
                    self, representation, key, value, **kwargs,
                )
            else:
                datasets[key] = TorchMusicFactoryDataset(
                    self, factory, key, value,  # type: ignore
                )

        return datasets

    def to_tensorflow_dataset(
        self,
        factory: Callable = None,
        representation: str = None,
        split_filename: Union[str, Path] = None,
        splits: Sequence[float] = None,
        random_state: Any = None,
        **kwargs: Any,
    ) -> Union["TFDataset", Dict[str, "TFDataset"]]:
        """Return the dataset as a TensorFlow dataset.

        Parameters
        ----------
        factory : Callable, optional
            Function to be applied to the Music objects. The input is a
            Music object, and the output is an array or a tensor.
        representation : str, optional
            Target representation. See :func:`muspy.to_representation()`
            for available representation.
        split_filename : str or Path, optional
            If given and exists, path to the file to read the split
            from. If None or not exists, path to save the split.
        splits : float or list of float, optional
            Ratios for train-test-validation splits. If None, return the
            full dataset as a whole. If float, return train and test
            splits. If list of two floats, return train and test splits.
            If list of three floats, return train, test and validation
            splits.
        random_state : int, array_like or RandomState, optional
            Random state used to create the splits. If int or
            array_like, the value is passed to
            :class:`numpy.random.RandomState`, and the created
            RandomState object is used to create the splits. If
            RandomState, it will be used to create the splits.

        Returns
        -------
        :class:tensorflow.data.Dataset` or Dict of
        :class:tensorflow.data.dataset`
            Converted TensorFlow dataset(s).

        """
        if representation is None and factory is None:
            raise TypeError(
                "One of `representation` and `factory` must be given."
            )
        if representation is not None and factory is not None:
            raise TypeError(
                "Only one of `representation` and `factory` can be given."
            )

        try:
            import tensorflow as tf
            from tensorflow.data import Dataset as TFDataset
        except ImportError:
            raise ImportError("Optional package tensorflow is required.")

        if representation is not None:

            def _gen(indices):
                for idx in indices:
                    yield self[idx].to_representation(representation, **kwargs)

        else:

            def _gen(indices):
                for idx in indices:
                    yield factory(self[idx])

        # TODO: `from_generator` is slow.

        # No split
        if splits is None:
            indices = np.arange(len(self))
            return TFDataset.from_generator(_gen, tf.float32, args=[indices])

        datasets: Dict[str, TFDataset] = {}
        indices_list = self.split(split_filename, splits, random_state)
        for key, value in indices_list.items():
            indices = np.array(value)
            datasets[key] = TFDataset.from_generator(
                _gen, tf.float32, args=[indices]
            )

        return datasets


class RemoteDataset(Dataset):
    """Base class for remote MusPy datasets.

    This class extends :class:`muspy.Dataset` to support remote
    datasets. To build a custom remote dataset, please refer to the
    documentation of :class:`muspy.Dataset` for details. In addition,
    set the class attribute ``_sources`` to the URLs to the source files
    (see Notes).

    Attributes
    ----------
    root : str or Path
        Root directory of the dataset.

    Parameters
    ----------
    download_and_extract : bool, default: False
        Whether to download and extract the dataset.
    overwrite : bool, default: False
        Whether to overwrite existing file(s).
    cleanup : bool, default: False
        Whether to remove the source archive(s).
    verbose : bool, default: True
        Whether to be verbose.

    Raises
    ------
    RuntimeError:
        If ``download_and_extract`` is False but file
        ``{root}/.muspy.success`` does not exist (see below).

    Important
    ---------
    :meth:`muspy.Dataset.exists` depends solely on a special file named
    ``.muspy.success`` in directory ``{root}/_converted/``. This file
    serves as an indicator for the existence and integrity of the
    dataset. It will automatically be created if the dataset is
    successfully downloaded and extracted by
    :meth:`muspy.Dataset.download_and_extract`. If the dataset is
    downloaded manually, make sure to create the ``.muspy.success`` file
    in directory ``{root}/_converted/`` to prevent errors.

    Notes
    -----
    The class attribute ``_sources`` is a dictionary storing the
    following information of each source file.

    - filename (str): Name to save the file.
    - url (str): URL to the file.
    - archive (bool): Whether the file is an archive.
    - md5 (str, optional): Expected MD5 checksum of the file.
    - sha256 (str, optional): Expected SHA256 checksum of the file.

    Here is an example.::

        _sources = {
            "example": {
                "filename": "example.tar.gz",
                "url": "https://www.example.com/example.tar.gz",
                "archive": True,
                "md5": None,
                "sha256": None,
            }
        }

    See Also
    --------
    :class:`muspy.Dataset` : Base class for MusPy datasets.

    """

    _sources: Dict[str, dict] = {}

    def __init__(
        self,
        root: Union[str, Path],
        download_and_extract: bool = False,
        overwrite: bool = False,
        cleanup: bool = False,
        verbose: bool = True,
    ):
        super().__init__()
        self.root = Path(root).expanduser().resolve()
        self.root.mkdir(exist_ok=True)

        if download_and_extract:
            self.download_and_extract(
                overwrite=overwrite, cleanup=cleanup, verbose=verbose
            )

        if not self.exists():
            raise RuntimeError(
                "Dataset not found. You can download it by passing "
                "`download_and_extract=True`."
            )

    def __repr__(self) -> str:
        return "{}(root={})".format(type(self).__name__, self.root)

    def __getitem__(self, index) -> Music:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError

    def exists(self) -> bool:
        """Return True if the dataset exists, otherwise False."""
        if not (self.root / ".muspy.success").is_file():
            return False
        return True

    def source_exists(self) -> bool:
        """Return True if all the sources exist, otherwise False."""
        for source in self._sources.values():
            filename = self.root / source["filename"]
            if not filename.is_file():
                return False
            if "size" in source and not check_size(filename, source["size"]):
                return False
            if "md5" in source and not check_md5(filename, source["md5"]):
                return False
            if "sha256" in source and not check_sha256(
                filename, source["sha256"]
            ):
                return False
        return True

    def download(
        self: RemoteDatasetType, overwrite: bool = False, verbose: bool = True
    ) -> RemoteDatasetType:
        """Download the dataset source(s).

        Parameters
        ----------
        overwrite : bool, default: False
            Whether to overwrite existing file(s).
        verbose : bool, default: True
            Whether to be verbose.

        Returns
        -------
        Object itself.

        """
        if self.exists():
            if verbose:
                print(
                    "Skip downloading as the `.muspy.success` file is found."
                )
            return self

        for source in self._sources.values():
            download_url(
                source["url"],
                self.root / source["filename"],
                overwrite=overwrite,
                size=source.get("size"),
                md5=source.get("md5"),
                sha256=source.get("sha256"),
                verbose=verbose,
            )
        return self

    def extract(
        self: RemoteDatasetType, cleanup: bool = False, verbose: bool = True
    ) -> RemoteDatasetType:
        """Extract the downloaded archive(s).

        Parameters
        ----------
        cleanup : bool, default: False
            Whether to remove the source archive after extraction.
        verbose : bool, default: True
            Whether to be verbose.

        Returns
        -------
        Object itself.

        """
        if self.exists():
            if verbose:
                print("Skip extracting as the `.muspy.success` file is found.")
            return self

        for source in self._sources.values():
            filename = self.root / source["filename"]
            if source["archive"]:
                extract_archive(
                    filename, self.root, cleanup=cleanup, verbose=verbose
                )
        (self.root / ".muspy.success").touch(exist_ok=True)
        return self

    def download_and_extract(
        self: RemoteDatasetType,
        overwrite: bool = False,
        cleanup: bool = False,
        verbose: bool = True,
    ) -> RemoteDatasetType:
        """Download source datasets and extract the downloaded archives.

        Parameters
        ----------
        overwrite : bool, default: False
            Whether to overwrite existing file(s).
        cleanup : bool, default: False
            Whether to remove the source archive(s).
        verbose : bool, default: True
            Whether to be verbose.

        Returns
        -------
        Object itself.

        """
        return self.download(overwrite=overwrite, verbose=verbose).extract(
            cleanup=cleanup, verbose=verbose
        )


def _get_filenames(root, extensions: List[str], recursive: bool = True):
    filenames = []
    for ext in extensions:
        if recursive:
            filenames.extend(root.rglob(f"*.{ext}"))
        else:
            filenames.extend(root.glob(f"*.{ext}"))
    return filenames


class MusicDataset(Dataset):
    """Class for datasets of MusPy JSON/YAML files.

    Parameters
    ----------
    root : str or Path
        Root directory of the dataset.
    kind : {'json', 'yaml'}, optional
        File formats to include in the dataset. Defaults to include
        both JSON and YAML files.

    Attributes
    ----------
    root : Path
        Root directory of the dataset.
    filenames : list of Path
        Path to the files, relative to `root`.

    See Also
    --------
    :class:`muspy.Dataset` : Base class for MusPy datasets.

    """

    def __init__(self, root: Union[str, Path], kind: str = None):
        if kind is not None and kind not in ("json", "yaml"):
            raise ValueError(f"Unknown value for `kind` : {kind} .")

        self.root = Path(root).expanduser().resolve()
        self.root.mkdir(exist_ok=True)

        if kind is None:
            extensions = ["json", "json.gz", "yaml", "yaml.gz"]
        elif kind == "json":
            extensions = ["json", "json.gz"]
        else:
            extensions = ["yaml", "yaml.gz"]
        self.filenames = _get_filenames(self.root, extensions)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(root={self.root})"

    def __getitem__(self, index) -> Music:
        return load(self.root / self.filenames[index])

    def __len__(self) -> int:
        return len(self.filenames)


class RemoteMusicDataset(MusicDataset, RemoteDataset):
    """Base class for remote datasets of MusPy JSON/YAML files.

    Parameters
    ----------
    root : str or Path
        Root directory of the dataset.
    download_and_extract : bool, default: False
        Whether to download and extract the dataset.
    overwrite : bool, default: False
        Whether to overwrite existing file(s).
    cleanup : bool, default: False
        Whether to remove the source archive(s).
    kind : {'json', 'yaml'}, optional
        File formats to include in the dataset. Defaults to include
        both JSON and YAML files.
    verbose : bool. default: True
        Whether to be verbose.

    Attributes
    ----------
    root : Path
        Root directory of the dataset.
    filenames : list of Path
        Path to the files, relative to `root`.

    See Also
    --------
    :class:`muspy.MusicDataset` :
        Class for datasets of MusPy JSON/YAML files.
    :class:`muspy.RemoteDataset` : Base class for remote MusPy datasets.

    """

    def __init__(
        self,
        root: Union[str, Path],
        download_and_extract: bool = False,
        overwrite: bool = False,
        cleanup: bool = False,
        kind: str = None,
        verbose: bool = True,
    ):
        RemoteDataset.__init__(
            self,
            root,
            download_and_extract=download_and_extract,
            overwrite=overwrite,
            cleanup=cleanup,
            verbose=verbose,
        )
        MusicDataset.__init__(self, root, kind=kind)


class FolderDataset(Dataset):
    """Class for datasets storing files in a folder.

    This class extends :class:`muspy.Dataset` to support folder
    datasets. To build a custom folder dataset, please refer to the
    documentation of :class:`muspy.Dataset` for details. In addition,
    set class attribute ``_extension`` to the extension to look for
    when building the dataset and set ``read`` to a callable that takes
    as inputs a filename of a source file and return the converted Music
    object.

    Attributes
    ----------
    root : str or Path
        Root directory of the dataset.

    Parameters
    ----------
    convert : bool, default: False
        Whether to convert the dataset to MusPy JSON/YAML files. If
        False, will check if converted data exists. If so, disable
        on-the-fly mode. If not, enable on-the-fly mode and warns.
    kind : {'json', 'yaml'}, default: 'json'
        File format to save the data.
    n_jobs : int, default: 1
        Maximum number of concurrently running jobs. If equal to 1,
        disable multiprocessing.
    ignore_exceptions : bool, default: True
        Whether to ignore errors and skip failed conversions. This can
        be helpful if some source files are known to be corrupted.
    use_converted : bool, optional
        Force to disable on-the-fly mode and use converted data.
        Defaults to True if converted data exist, otherwise False.

    Important
    ---------
    :meth:`muspy.FolderDataset.converted_exists` depends solely on a
    special file named ``.muspy.success`` in the folder
    ``{root}/_converted/``, which serves as an indicator for the
    existence and integrity of the converted dataset. If the converted
    dataset is built by :meth:`muspy.FolderDataset.convert`, the
    ``.muspy.success`` file will be created as well. If the converted
    dataset is created manually, make sure to create the
    ``.muspy.success`` file in the folder ``{root}/_converted/`` to
    prevent errors.

    Notes
    -----
    Two modes are available for this dataset. When the on-the-fly mode
    is enabled, a data sample is converted to a music object on the fly
    when being indexed. When the on-the-fly mode is disabled, a data
    sample is loaded from the precomputed converted data.

    See Also
    --------
    :class:`muspy.Dataset` : Base class for MusPy datasets.

    """

    _extension: str = ""

    def __init__(
        self,
        root: Union[str, Path],
        convert: bool = False,
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        use_converted: bool = None,
    ):
        self.root = Path(root).expanduser().resolve()
        self.kind = kind

        # A pointer to the callable used to produce the Music object
        self._factory: Callable = lambda: None

        # A pointer to the list of filenames used when indexing
        self._filenames: list = []

        self.raw_filenames: list = []
        self.converted_filenames: list = []

        if convert:
            self.convert(kind, n_jobs, ignore_exceptions)

        if use_converted is None:
            use_converted = self.converted_exists()

        if use_converted:
            self.use_converted()
        else:
            self.on_the_fly()

        if not self._filenames:
            raise ValueError("Nothing found in the directory.")

        (self.root / ".muspy.success").touch()

    @property
    def converted_dir(self):
        """Path to the root directory of the converted dataset."""
        return self.root / "_converted"

    def __repr__(self) -> str:
        return f"{type(self).__name__}(root={self.root})"

    def __getitem__(self, index) -> Music:
        return self._factory(self._filenames[index])

    def __len__(self) -> int:
        return len(self._filenames)

    def read(self, filename: Any) -> Music:
        """Read a file into a Music object."""
        raise NotImplementedError

    def load(self, filename: Union[str, Path]) -> Music:
        """Load a file into a Music object."""
        return load(self.root / filename)

    def exists(self) -> bool:
        """Return True if the dataset exists, otherwise False."""
        if not (self.root / ".muspy.success").is_file():
            return False
        return True

    def converted_exists(self) -> bool:
        """Return True if the saved dataset exists, otherwise False."""
        if not (self.converted_dir / ".muspy.success").is_file():
            return False
        return True

    def get_converted_filenames(self):
        """Return a list of converted filenames."""
        return sorted(self.converted_dir.rglob("*." + self.kind))

    def use_converted(self: FolderDatasetType) -> FolderDatasetType:
        """Disable on-the-fly mode and use converted data.

        Returns
        -------
        Object itself.

        """
        if not self.converted_exists():
            raise RuntimeError(
                "Converted data not found. Run `convert()` to convert "
                "the dataset."
            )
        if not self.converted_filenames:
            self.converted_filenames = self.get_converted_filenames()
        self._filenames = self.converted_filenames
        self._use_converted = True
        self._factory = self.load
        return self

    def get_raw_filenames(self):
        """Return a list of raw filenames."""
        return sorted(
            (
                filename
                for filename in self.root.rglob("*." + self._extension)
                if not str(filename.relative_to(self.root)).startswith(
                    "_converted/"
                )
            )
        )

    def on_the_fly(self: FolderDatasetType) -> FolderDatasetType:
        """Enable on-the-fly mode and convert the data on the fly.

        Returns
        -------
        Object itself.

        """
        if not self.raw_filenames:
            self.raw_filenames = self.get_raw_filenames()
        self._filenames = self.raw_filenames
        self._use_converted = False
        self._factory = self.read
        return self

    def convert(
        self: FolderDatasetType,
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        verbose: bool = True,
        **kwargs,
    ) -> FolderDatasetType:
        """Convert and save the Music objects.

        The converted files will be named by its index and saved to
        ``root/_converted``. The original filenames can be found in the
        ``filenames`` attribute. For example, the file at
        ``filenames[i]`` will be converted and saved to ``{i}.json``.

        Parameters
        ----------
        kind : {'json', 'yaml'}, default: 'json'
            File format to save the data.
        n_jobs : int, default: 1
            Maximum number of concurrently running jobs. If equal to 1,
            disable multiprocessing.
        ignore_exceptions : bool, default: True
            Whether to ignore errors and skip failed conversions. This
            can be helpful if some source files are known to be
            corrupted.
        verbose : bool, default: True
            Whether to be verbose.
        **kwargs
            Keyword arguments to pass to :func:`muspy.save`.

        Returns
        -------
        Object itself.

        """
        if self.converted_exists():
            if verbose:
                print("Skip conversion as the `.muspy.success` file is found.")
            return self
        self.on_the_fly()
        self.converted_dir.mkdir(exist_ok=True)
        self.save(
            self.converted_dir,
            kind=kind,
            n_jobs=n_jobs,
            ignore_exceptions=ignore_exceptions,
            verbose=verbose,
            **kwargs,
        )
        (self.converted_dir / ".muspy.success").touch(exist_ok=True)
        self.use_converted()
        self.kind = kind
        return self


class RemoteFolderDataset(FolderDataset, RemoteDataset):
    """Base class for remote datasets storing files in a folder.

    Attributes
    ----------
    root : str or Path
        Root directory of the dataset.

    Parameters
    ----------
    download_and_extract : bool, default: False
        Whether to download and extract the dataset.
    cleanup : bool, default: False
        Whether to remove the source archive(s).
    convert : bool, default: False
        Whether to convert the dataset to MusPy JSON/YAML files. If
        False, will check if converted data exists. If so, disable
        on-the-fly mode. If not, enable on-the-fly mode and warns.
    kind : {'json', 'yaml'}, default: 'json'
        File format to save the data.
    n_jobs : int, default: 1
        Maximum number of concurrently running jobs. If equal to 1,
        disable multiprocessing.
    ignore_exceptions : bool, default: True
        Whether to ignore errors and skip failed conversions. This can
        be helpful if some source files are known to be corrupted.
    use_converted : bool, optional
        Force to disable on-the-fly mode and use converted data.
        Defaults to True if converted data exist, otherwise False.

    See Also
    --------
    :class:`muspy.FolderDataset` :
        Class for datasets storing files in a folder.
    :class:`muspy.RemoteDataset` : Base class for remote MusPy datasets.

    """

    def __init__(
        self,
        root: Union[str, Path],
        download_and_extract: bool = False,
        overwrite: bool = False,
        cleanup: bool = False,
        convert: bool = False,
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        use_converted: bool = None,
        verbose: bool = True,
    ):
        RemoteDataset.__init__(
            self,
            root,
            download_and_extract=download_and_extract,
            overwrite=overwrite,
            cleanup=cleanup,
            verbose=verbose,
        )
        FolderDataset.__init__(
            self,
            root,
            convert=convert,
            kind=kind,
            n_jobs=n_jobs,
            ignore_exceptions=ignore_exceptions,
            use_converted=use_converted,
        )

    def read(self, filename: str) -> Music:
        """Read a file into a Music object."""
        raise NotImplementedError


class ABCFolderDataset(FolderDataset):
    """Class for datasets storing ABC files in a folder.

    See Also
    --------
    :class:`muspy.FolderDataset` :
        Class for datasets storing files in a folder.

    """

    _extension = "abc"

    def read(self, filename: Tuple[str, Tuple[int, int]]) -> Music:
        """Read a file into a Music object."""
        filename_, (start, end) = filename
        data = []
        with open(filename_) as f:
            for idx, line in enumerate(f):
                if start <= idx < end and not line.startswith("%"):
                    data.append(line)
        return read_abc_string("".join(data))  # type: ignore

    def on_the_fly(self: FolderDatasetType) -> FolderDatasetType:
        """Enable on-the-fly mode and convert the data on the fly.

        Returns
        -------
        Object itself.

        """
        if not self.raw_filenames:
            filenames = sorted(
                (
                    filename
                    for filename in self.root.rglob("*." + self._extension)
                    if not str(filename.relative_to(self.root)).startswith(
                        "_converted/"
                    )
                )
            )
            self.raw_filenames = []
            for filename in filenames:
                idx = 0
                start = 0
                with open(filename, errors="ignore") as f:

                    # Detect parts in a file
                    for idx, line in enumerate(f):
                        if line.startswith("X:"):
                            if start:
                                self.raw_filenames.append(
                                    (filename, (start, idx))
                                )
                            start = idx

                    # Append the last part
                    if start:
                        self.raw_filenames.append((filename, (start, idx)))

        self._filenames = self.raw_filenames
        self._use_converted = False
        self._factory = self.read
        return self


class RemoteABCFolderDataset(ABCFolderDataset, RemoteDataset):
    """Base class for remote datasets storing ABC files in a folder.

    See Also
    --------
    :class:`muspy.ABCFolderDataset` :
        Class for datasets storing ABC files in a folder.
    :class:`muspy.RemoteDataset` : Base class for remote MusPy datasets.

    """

    def __init__(
        self,
        root: Union[str, Path],
        download_and_extract: bool = False,
        overwrite: bool = False,
        cleanup: bool = False,
        convert: bool = False,
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        use_converted: bool = None,
        verbose: bool = True,
    ):
        RemoteDataset.__init__(
            self,
            root,
            download_and_extract=download_and_extract,
            overwrite=overwrite,
            cleanup=cleanup,
            verbose=verbose,
        )
        ABCFolderDataset.__init__(
            self,
            root,
            convert=convert,
            kind=kind,
            n_jobs=n_jobs,
            ignore_exceptions=ignore_exceptions,
            use_converted=use_converted,
        )



"""Setup script."""
from pathlib import Path

from setuptools import find_packages, setup


def _get_long_description():
    with open(str(Path(__file__).parent / "README.md"), "r") as f:
        return f.read()


def _get_version():
    with open(str(Path(__file__).parent / "muspy/version.py"), "r") as f:
        for line in f:
            if line.startswith("__version__"):
                delimeter = '"' if '"' in line else "'"
                return line.split(delimeter)[1]
    raise RuntimeError("Cannot read version string.")


VERSION = _get_version()

setup(
    name="muspy",
    version=VERSION,
    author="Hao-Wen Dong",
    author_email="salu.hwdong@gmail.com",
    description="A toolkit for symbolic music generation",
    long_description=_get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/salu133445/muspy",
    download_url=(
        f"https://github.com/salu133445/muspy/archive/v{VERSION}.tar.gz"
    ),
    project_urls={"Documentation": "https://salu133445.github.io/muspy/"},
    license="MIT",
    keywords=[
        "music",
        "audio",
        "music-generation",
        "music-information-retrieval",
    ],
    packages=find_packages(include=["muspy", "muspy.*"], exclude=["tests"]),
    install_requires=[
        "bidict>=0.21",
        "PyYAML>=3.0",
        "joblib>=0.15",
        "matplotlib>=1.5",
        "mido>=1.0",
        "music21>=5.0",
        "pretty-midi>=0.2",
        "pypianoroll>=1.0",
        "requests>=2.0",
        "tqdm>=4.0",
    ],
    extras_require={
        "dev": [
            "black>=19.0",
            "flake8-docstrings>=1.5",
            "flake8>=3.8",
            "mypy>=0.900",
            "pylint>=2.5",
            "sphinx-rtd-theme>=0.5",
            "sphinx>=3.0",
        ],
        "optional": ["tensorflow>=2.0", "torch>=1.0"],
        "schema": ["jsonschema>=3.0", "xmlschema>=1.0", "yamale>=2.0"],
        "test": ["pytest>=6.0", "pytest-cov>=2.0"],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Multimedia :: Sound/Audio",
    ],
    python_requires=">=3.6",
)



"""JSON input interface."""
import gzip
import json
from pathlib import Path
from typing import TextIO, Union

from ..music import Music


def load_json(
    path: Union[str, Path, TextIO], compressed: bool = None
) -> Music:
    """Load a JSON file into a Music object.

    Parameters
    ----------
    path : str, Path or TextIO
        Path to the file or the file to load.
    compressed : bool, optional
        Whether the file is a compressed JSON file (`.json.gz`). Has no
        effect when `path` is a file object. Defaults to infer from the
        extension (`.gz`).

    Returns
    -------
    :class:`muspy.Music`
        Loaded Music object.

    Notes
    -----
    When a path is given, assume UTF-8 encoding and gzip compression if
    `compressed=True`.

    """
    if isinstance(path, (str, Path)):
        if compressed is None:
            if str(path).lower().endswith(".gz"):
                compressed = True
            else:
                compressed = False
        if compressed:
            with gzip.open(path, "rt", encoding="utf-8") as f:
                return Music.from_dict(json.load(f))
        with open(path, encoding="utf-8") as f:
            return Music.from_dict(json.load(f))

    return Music.from_dict(json.load(path))



"""Hymnal Dataset."""
from pathlib import Path
from typing import Union

import requests

from ..inputs import read_midi
from ..music import Music
from .base import DatasetInfo, FolderDataset

_NAME = "Hymnal Dataset"
_DESCRIPTION = """\
The Hymnal Dataset is a collection of hymns in MIDI format available at
hymnal.net."""
_HOMEPAGE = "https://www.hymnal.net/"


class HymnalDataset(FolderDataset):
    """Hymnal Dataset."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _extension = "mid"
    _type = "mid"

    def __init__(
        self,
        root: Union[str, Path],
        download: bool = False,
        convert: bool = False,
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        use_converted: bool = None,
    ):
        self.root = Path(root).expanduser().resolve()
        if not self.root.exists():
            raise ValueError("`root` must be an existing path.")
        if not self.root.is_dir():
            raise ValueError("`root` must be a directory.")

        if download:
            self.download()

        super().__init__(
            root, convert, kind, n_jobs, ignore_exceptions, use_converted
        )

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)

    def download(self) -> "FolderDataset":
        """Download the source datasets.

        Returns
        -------
        Object itself.

        """
        # Maximum consecutive trials allowed to fail
        tolerance = 10

        kinds = ["Classic", "New Tunes", "New Songs", "Children"]
        keys = ["h", "nt", "ns", "c"]

        print("Downloading sources.")
        for kind, key in zip(kinds, keys):
            # Make sure the folder exists
            (self.root / kind).mkdir(exist_ok=True)

            # Reset the index and the consecutive failure counter
            idx = 1
            consecutive_failure_count = 0

            # Loop until the number of consecutive failures exceed
            # the tolerance
            while consecutive_failure_count < tolerance:
                # Send a HEAD request to check if content type is MIDI
                url = (
                    f"https://www.hymnal.net/en/hymn/{key}/{idx}/"
                    f"f={self._type}"
                )
                req = requests.head(url)
                if req.headers["Content-Type"] != "audio/midi":
                    consecutive_failure_count += 1
                    continue

                # Send another HEAD request to check if we have
                # exceeded the total number of pieces -> When we request
                # for an out of bound index, it seems that it will
                # randomly return another piece. Thus, if the first and
                # the second requests have different content sizes, we
                # can break the loop.
                second_req = requests.head(url)
                if (
                    second_req.headers["Content-Length"]
                    != req.headers["Content-Length"]
                ):
                    break

                # Send a GET request to get the MIDI file
                req = requests.get(url)
                filename = str(self.root / kind / (str(idx) + ".mid"))
                with open(filename, "wb") as f:
                    f.write(req.content)

                # Reset the consecutive failure counter
                if consecutive_failure_count:
                    consecutive_failure_count = 0

                idx += 1

                if idx % 100 == 0:
                    print(f"Successfully downloaded {idx} files.")

        (self.root / ".muspy.success").touch(exist_ok=True)
        return self


class HymnalTuneDataset(FolderDataset):
    """Hymnal Dataset (tune only)."""

    _info = DatasetInfo(_NAME, _DESCRIPTION, _HOMEPAGE)
    _extension = "mid"
    _type = "tune"

    def __init__(
        self,
        root: Union[str, Path],
        download: bool = False,
        convert: bool = False,
        kind: str = "json",
        n_jobs: int = 1,
        ignore_exceptions: bool = True,
        use_converted: bool = None,
    ):

        self.root = Path(root).expanduser().resolve()
        if not self.root.exists():
            raise ValueError("`root` must be an existing path.")
        if not self.root.is_dir():
            raise ValueError("`root` must be a directory.")

        if download:
            self.download()

        super().__init__(
            root, convert, kind, n_jobs, ignore_exceptions, use_converted
        )

    def read(self, filename: Union[str, Path]) -> Music:
        """Read a file into a Music object."""
        return read_midi(self.root / filename)

    def download(self) -> "FolderDataset":
        """Download the source datasets.

        Returns
        -------
        Object itself.

        """
        # Maximum consecutive trials allowed to fail
        tolerance = 10

        kinds = ["Classic", "New Tunes", "New Songs", "Children"]
        keys = ["h", "nt", "ns", "c"]

        print("Downloading sources.")
        for kind, key in zip(kinds, keys):
            # Make sure the folder exists
            (self.root / kind).mkdir(exist_ok=True)

            # Reset the index and the consecutive failure counter
            idx = 1
            consecutive_failure_count = 0

            # Loop until the number of consecutive failures exceed
            # the tolerance
            while consecutive_failure_count < tolerance:
                # Send a HEAD request to check if content type is MIDI
                url = (
                    f"https://www.hymnal.net/en/hymn/{key}/{idx}/"
                    f"f={self._type}"
                )
                req = requests.head(url)
                if req.headers["Content-Type"] != "audio/midi":
                    consecutive_failure_count += 1
                    continue

                # Send another HEAD request to check if we have
                # exceeded the total number of pieces -> When we request
                # for an out of bound index, it seems that it will
                # randomly return another piece. Thus, if the first and
                # the second requests have different content sizes, we
                # can break the loop.
                second_req = requests.head(url)
                if (
                    second_req.headers["Content-Length"]
                    != req.headers["Content-Length"]
                ):
                    break

                # Send a GET request to get the MIDI file
                req = requests.get(url)
                filename = str(self.root / kind / (str(idx) + ".mid"))
                with open(filename, "wb") as f:
                    f.write(req.content)

                # Reset the consecutive failure counter
                if consecutive_failure_count:
                    consecutive_failure_count = 0

                idx += 1

                if idx % 100 == 0:
                    print(f"Successfully downloaded {idx} files.")

        (self.root / ".muspy.success").touch(exist_ok=True)
        return self
